""" Crossover implementations for GA.
"""

# Author: Genevieve Hayes (modified by Andrew Rollings)
# License: BSD 3 clause

import numpy as np

from mlrose_hiive.algorithms.crossovers._crossover_base import _CrossOverBase


class UniformCrossOver(_CrossOverBase):
    def __init__(self, opt_prob):
        super().__init__(opt_prob)

    def mate(self, p1, p2):
        n = np.random.randint(2, size=self._length)
        x = np.vstack((p1, p2))
        child = x[n, np.arange(self._length)]
        return child
    """
    def mate(self, p1, p2):
        n = np.random.choice(a=[False, True], size=self._length)
        child = np.array([p1[i] if n[i] else p2[i] for i in range(self._length)])
        return child
    """




""" Classes for defining optimization problem objects."""

# Author: Genevieve Hayes (modified by Andrew Rollings)
# License: BSD 3 clause

from mlrose_hiive.algorithms.crossovers import UniformCrossOver
from mlrose_hiive.algorithms.mutators import ChangeOneMutator
from mlrose_hiive.fitness.knapsack import Knapsack
from mlrose_hiive.opt_probs.discrete_opt import DiscreteOpt


class KnapsackOpt(DiscreteOpt):
    def __init__(self, length=None, fitness_fn=None, maximize=True, max_val=2,
                 weights=None, values=None, max_weight_pct=0.35,
                 crossover=None, mutator=None,
                 multiply_by_max_item_count=False):

        if (fitness_fn is None) and (weights is None and values is None):
            raise Exception("""fitness_fn or both weights and"""
                            + """ values must be specified.""")

        if length is None:
            if weights is not None:
                length = len(weights)
            elif values is not None:
                length = len(values)
            elif fitness_fn is not None:
                length = len(fitness_fn.weights)

        self.length = length

        if fitness_fn is None:
            fitness_fn = Knapsack(weights=weights, values=values,
                                  max_weight_pct=max_weight_pct,
                                  max_item_count=max_val,
                                  multiply_by_max_item_count=multiply_by_max_item_count)

        self.max_val = max_val
        crossover = UniformCrossOver(self) if crossover is None else crossover
        mutator = ChangeOneMutator(self) if mutator is None else mutator
        super().__init__(length, fitness_fn, maximize, max_val, crossover, mutator)



""" Functions to implement the randomized optimization and search algorithms.
"""

# Author: Genevieve Hayes (modified by Andrew Rollings)
# License: BSD 3 clause

import numpy as np

from mlrose_hiive.decorators import short_name


def _get_hamming_distance_default(population, p1):
    hamming_distances = np.array([np.count_nonzero(p1 != p2) / len(p1) for p2 in population])
    return hamming_distances


def _get_hamming_distance_float(population, p1):
    # use squares instead?
    hamming_distances = np.array([np.abs(np.diff(p1, p2)) / len(p1) for p2 in population])
    return hamming_distances


def _genetic_alg_select_parents(pop_size, problem,
                                get_hamming_distance_func,
                                hamming_factor=0.0):
    mating_probabilities = problem.get_mate_probs()
    if get_hamming_distance_func is not None and hamming_factor > 0.01:
        selected = np.random.choice(pop_size, p=mating_probabilities)
        population = problem.get_population()
        p1 = population[selected]
        hamming_distances = get_hamming_distance_func(population, p1)
        hfa = hamming_factor / (1.0 - hamming_factor)
        hamming_distances = (hamming_distances * hfa) * mating_probabilities
        hamming_distances /= hamming_distances.sum()
        selected = np.random.choice(pop_size, p=hamming_distances)
        p2 = population[selected]

        return p1, p2

    selected = np.random.choice(pop_size,
                                size=2,
                                p=mating_probabilities)
    p1 = problem.get_population()[selected[0]]
    p2 = problem.get_population()[selected[1]]
    return p1, p2


@short_name('ga')
def genetic_alg(problem, pop_size=200, pop_breed_percent=0.75, elite_dreg_ratio=0.99,
                minimum_elites=0, minimum_dregs=0, mutation_prob=0.1,
                max_attempts=10, max_iters=np.inf, curve=False, random_state=None,
                state_fitness_callback=None, callback_user_info=None,
                hamming_factor=0.0, hamming_decay_factor=None):
    """Use a standard genetic algorithm to find the optimum for a given
    optimization problem.
    Parameters
    ----------
    problem: optimization object
        Object containing fitness function optimization problem to be solved.
        For example, :code:`DiscreteOpt()`, :code:`ContinuousOpt()` or
    pop_size: int, default: 200
        Size of population to be used in genetic algorithm.
    pop_breed_percent: float, default 0.75
        Percentage of population to breed in each iteration.
        The remainder of the population will be filled from the elite and
        dregs of the prior generation in a ratio specified by elite_dreg_ratio.
    elite_dreg_ratio: float, default:0.95
        The ratio of elites:dregs added directly to the next generation.
        For the default value, 95% of the added population will be elites,
        5% will be dregs.
    minimum_elites: int, default: 0
        Minimum number of elites to be added to next generation
    minimum_dregs: int, default: 0
        Minimum number of dregs to be added to next generation
    mutation_prob: float, default: 0.1
        Probability of a mutation at each element of the state vector
        during reproduction, expressed as a value between 0 and 1.
    max_attempts: int, default: 10
        Maximum number of attempts to find a better state at each step.
    max_iters: int, default: np.inf
        Maximum number of iterations of the algorithm.
    curve: bool, default: False
        Boolean to keep fitness values for a curve.
        If :code:`False`, then no curve is stored.
        If :code:`True`, then a history of fitness values is provided as a
        third return value.
    random_state: int, default: None
        If random_state is a positive integer, random_state is the seed used
        by np.random.seed(); otherwise, the random seed is not set.
    state_fitness_callback: function taking five parameters, default: None
        If specified, this callback will be invoked once per iteration.
        Parameters are (iteration, max attempts reached?, current best state, current best fit, user callback data).
        Return true to continue iterating, or false to stop.
    callback_user_info: any, default: None
        User data passed as last parameter of callback.
    Returns
    -------
    best_state: array
        Numpy array containing state that optimizes the fitness function.
    best_fitness: float
        Value of fitness function at best state.
    fitness_curve: array
        Numpy array of arrays containing the fitness of the entire population
        at every iteration.
        Only returned if input argument :code:`curve` is :code:`True`.
    References
    ----------
    Russell, S. and P. Norvig (2010). *Artificial Intelligence: A Modern
    Approach*, 3rd edition. Prentice Hall, New Jersey, USA.
    """
    if pop_size < 0:
        raise Exception("""pop_size must be a positive integer.""")
    elif not isinstance(pop_size, int):
        if pop_size.is_integer():
            pop_size = int(pop_size)
        else:
            raise Exception("""pop_size must be a positive integer.""")

    breeding_pop_size = int(pop_size * pop_breed_percent) - (minimum_elites + minimum_dregs)
    if breeding_pop_size < 1:
        raise Exception("""pop_breed_percent must be large enough to ensure at least one mating.""")

    if pop_breed_percent > 1:
        raise Exception("""pop_breed_percent must be less than 1.""")

    if (elite_dreg_ratio < 0) or (elite_dreg_ratio > 1):
        raise Exception("""elite_dreg_ratio must be between 0 and 1.""")

    if (mutation_prob < 0) or (mutation_prob > 1):
        raise Exception("""mutation_prob must be between 0 and 1.""")

    if (not isinstance(max_attempts, int) and not max_attempts.is_integer()) \
       or (max_attempts < 0):
        raise Exception("""max_attempts must be a positive integer.""")

    if (not isinstance(max_iters, int) and max_iters != np.inf
            and not max_iters.is_integer()) or (max_iters < 0):
        raise Exception("""max_iters must be a positive integer.""")

    # Set random seed
    if isinstance(random_state, int) and random_state > 0:
        np.random.seed(random_state)

    fitness_curve = []

    # Initialize problem, population and attempts counter
    problem.reset()
    problem.random_pop(pop_size)
    if state_fitness_callback is not None:
        # initial call with base data
        state_fitness_callback(iteration=0,
                               state=problem.get_state(),
                               fitness=problem.get_adjusted_fitness(),
                               fitness_evaluations=problem.fitness_evaluations,
                               user_data=callback_user_info)
    # check for hamming
    # get_hamming_distance_default_

    get_hamming_distance_func = None
    if hamming_factor > 0:
        g1 = problem.get_population()[0][0]
        if isinstance(g1, float) or g1.dtype == 'float64':
            get_hamming_distance_func = _get_hamming_distance_float
        else:
            get_hamming_distance_func = _get_hamming_distance_default

    attempts = 0
    iters = 0

    # initialize survivor count, elite count and dreg count
    survivors_size = pop_size - breeding_pop_size
    dregs_size = max(int(survivors_size * (1.0 - elite_dreg_ratio)) if survivors_size > 1 else 0, minimum_dregs)
    elites_size = max(survivors_size - dregs_size, minimum_elites)
    if dregs_size + elites_size > survivors_size:
        over_population = dregs_size + elites_size - survivors_size
        breeding_pop_size -= over_population

    continue_iterating = True
    while (attempts < max_attempts) and (iters < max_iters):
        iters += 1
        problem.current_iteration += 1

        # Calculate breeding probabilities
        problem.eval_mate_probs()

        # Create next generation of population
        next_gen = []
        for _ in range(breeding_pop_size):
            # Select parents
            parent_1, parent_2 = _genetic_alg_select_parents(pop_size=pop_size,
                                                             problem=problem,
                                                             hamming_factor=hamming_factor,
                                                             get_hamming_distance_func=get_hamming_distance_func)

            # Create offspring
            child = problem.reproduce(parent_1, parent_2, mutation_prob)
            next_gen.append(child)

        # fill remaining population with elites/dregs
        if survivors_size > 0:
            last_gen = list(zip(problem.get_population(), problem.get_pop_fitness()))
            sorted_parents = sorted(last_gen, key=lambda f: -f[1])
            best_parents = sorted_parents[:elites_size]
            next_gen.extend([p[0] for p in best_parents])
            if dregs_size > 0:
                worst_parents = sorted_parents[-dregs_size:]
                next_gen.extend([p[0] for p in worst_parents])

        next_gen = np.array(next_gen[:pop_size])
        problem.set_population(next_gen)

        next_state = problem.best_child()
        next_fitness = problem.eval_fitness(next_state)

        # If best child is an improvement,
        # move to that state and reset attempts counter
        current_fitness = problem.get_fitness()
        if next_fitness > current_fitness:
            problem.set_state(next_state)
            attempts = 0
        else:
            attempts += 1

        if curve:
            fitness_curve.append(problem.get_adjusted_fitness())

        # invoke callback
        if state_fitness_callback is not None:
            max_attempts_reached = (attempts == max_attempts) or (iters == max_iters) or problem.can_stop()
            continue_iterating = state_fitness_callback(iteration=iters,
                                                        attempt=attempts + 1,
                                                        done=max_attempts_reached,
                                                        state=problem.get_state(),
                                                        fitness=problem.get_adjusted_fitness(),
                                                        fitness_evaluations=problem.fitness_evaluations,
                                                        curve=np.asarray(fitness_curve) if curve else None,
                                                        user_data=callback_user_info)

        # decay hamming factor
        if hamming_decay_factor is not None and hamming_factor > 0.0:
            hamming_factor *= hamming_decay_factor
            hamming_factor = max(min(hamming_factor, 1.0), 0.0)
        # print(hamming_factor)

        # break out if requested
        if not continue_iterating:
            break
    best_fitness = problem.get_maximize()*problem.get_fitness()
    best_state = problem.get_state()
    return best_state, best_fitness, np.asarray(fitness_curve) if curve else None



""" GA Mutators."""

# Author: Genevieve Hayes
# Modified: Andrew Rollings
# License: BSD 3 clause
import numpy as np

from mlrose_hiive.algorithms.mutators._mutator_base import _MutatorBase


class DiscreteMutator(_MutatorBase):

    def __init__(self, opt_prob):
        super().__init__(opt_prob)
        self._max_val = opt_prob.max_val

    def mutate(self, child, mutation_probability):
        rand = np.random.uniform(size=self._length)
        mutate = np.where(rand < mutation_probability)[0]

        if self._max_val == 2:
            for i in mutate:
                child[i] = np.abs(child[i] - 1)

        else:
            for i in mutate:
                vals = list(np.arange(self._max_val))
                vals.remove(child[i])
                child[i] = vals[np.random.randint(0, self._max_val-1)]
        return child



import mlrose_hiive
from mlrose_hiive.decorators import short_name
from mlrose_hiive.runners._runner_base import _RunnerBase
import numpy as np

"""
Example usage:

    experiment_name = 'example_experiment'
    problem = TSPGenerator.generate(seed=SEED, number_of_cities=22)
    
    # note that you can also initialize a temperature_list this way
    # temperature_list = [mlrose_hiive.GeomDecay(init_temp=t, decay=d) for (t, d) in [(1, 0.99), (1e2, 0.999)]]
    # if you use this form, the decay_list parameter is ignored.

    sa = SARunner(problem=problem,
                  experiment_name=experiment_name,
                  output_directory=OUTPUT_DIRECTORY,
                  seed=SEED,
                  iteration_list=2 ** np.arange(14),
                  max_attempts=5000,
                  temperature_list=[1, 10, 50, 100, 250, 500, 1000, 2500, 5000, 10000],
                  decay_list=[mlrose_hiive.GeomDecay])

    # the two data frames will contain the results
    df_run_stats, df_run_curves = sa.run()                  
"""


@short_name('sa')
class SARunner(_RunnerBase):

    def __init__(self, problem, experiment_name, seed, iteration_list, temperature_list, decay_list=None,
                 max_attempts=500, generate_curves=True, **kwargs):
        super().__init__(problem=problem, experiment_name=experiment_name, seed=seed, iteration_list=iteration_list,
                         max_attempts=max_attempts, generate_curves=generate_curves,
                         **kwargs)
        self.use_raw_temp = True
        self.temperature_list = temperature_list
        if all([np.isscalar(x) for x in temperature_list]):
            if decay_list is None:
                decay_list = [mlrose_hiive.GeomDecay]
            self.decay_list = decay_list
            self.use_raw_temp = False

    def run(self):
        temperatures = self.temperature_list if self.use_raw_temp else [d(init_temp=t) for t in self.temperature_list
                                                                        for d in self.decay_list]
        return super().run_experiment_(algorithm=mlrose_hiive.simulated_annealing,
                                       schedule=('Temperature', temperatures))



import sklearn.metrics as skmt

from mlrose_hiive import NNClassifier, relu
from mlrose_hiive.decorators import short_name
from mlrose_hiive.runners._nn_runner_base import _NNRunnerBase
from mlrose_hiive.decorators import get_short_name

"""
Example usage:
    from mlrose_hiive.runners import NNGSRunner

    grid_search_parameters = ({
        'max_iters': [1, 2, 4, 8, 16, 32, 64, 128],                     # nn params
        'learning_rate': [0.001, 0.002, 0.003],                         # nn params
        'schedule': [ArithDecay(1), ArithDecay(100), ArithDecay(1000)]  # sa params
    })

    nnr = NNGSRunner(x_train=x_train,
                     y_train=y_train,
                     x_test=x_test,
                     y_test=y_test,
                     experiment_name='nn_test',
                     algorithm=mlrose_hiive.algorithms.sa.simulated_annealing,
                     grid_search_parameters=grid_search_parameters,
                     iteration_list=[1, 10, 50, 100, 250, 500, 1000, 2500, 5000, 10000],
                     hidden_layer_sizes=[[44,44]],
                     bias=True,
                     early_stopping=False,
                     clip_max=1e+10,
                     max_attempts=500,
                     generate_curves=True,
                     seed=200972)

    results = nnr.run()          # GridSearchCV instance returned    
"""


@short_name('nngs')
class NNGSRunner(_NNRunnerBase):

    def __init__(self, x_train, y_train, x_test, y_test, experiment_name, seed, iteration_list, algorithm,
                 grid_search_parameters, grid_search_scorer_method=skmt.balanced_accuracy_score,
                 bias=True, early_stopping=True, clip_max=1e+10, activation=None,
                 max_attempts=500, n_jobs=1, cv=5, generate_curves=True, output_directory=None,
                 **kwargs):

        # update short name based on algorithm
        self._set_dynamic_runner_name(f'{get_short_name(self)}_{get_short_name(algorithm)}')

        # take a copy of the grid search parameters
        grid_search_parameters = {**grid_search_parameters}

        # hack for compatibility purposes
        if 'max_iter' in grid_search_parameters:
            grid_search_parameters['max_iter'] = grid_search_parameters.pop('max_iters')

        # call base class init
        super().__init__(x_train=x_train, y_train=y_train, x_test=x_test, y_test=y_test,
                         experiment_name=experiment_name,
                         seed=seed,
                         iteration_list=iteration_list,
                         grid_search_parameters=grid_search_parameters,
                         generate_curves=generate_curves,
                         output_directory=output_directory,
                         n_jobs=n_jobs,
                         cv=cv,
                         grid_search_scorer_method=grid_search_scorer_method,
                         **kwargs)

        # build the classifier
        self.classifier = NNClassifier(runner=self,
                                       algorithm=algorithm,
                                       max_attempts=max_attempts,
                                       clip_max=clip_max,
                                       early_stopping=early_stopping,
                                       seed=seed,
                                       bias=bias)

    def run_one_experiment_(self, algorithm, total_args, **params):
        if self._extra_args is not None and len(self._extra_args) > 0:
            params = {**params, **self._extra_args}

        total_args.update(params)
        total_args.pop('problem')
        user_info = [(k, v) for k, v in total_args.items()]

        return self._invoke_algorithm(algorithm=algorithm,
                                      curve=self.generate_curves,
                                      user_info=user_info,
                                      additional_algorithm_args=total_args,
                                      **params)



""" Classes for defining optimization problem objects."""

# Author: Andrew Rollings
# License: BSD 3 clause

import numpy as np

from mlrose_hiive import FlipFlopOpt


class FlipFlopGenerator:
    @staticmethod
    def generate(seed, size=20):
        np.random.seed(seed)
        problem = FlipFlopOpt(length=size)
        return problem



""" GA Mutators."""

# Author: Genevieve Hayes
# Modified: Andrew Rollings
# License: BSD 3 clause

from .change_one_mutator import ChangeOneMutator
from .discrete_mutator import DiscreteMutator
from .swap_mutator import SwapMutator
from .shift_one_mutator import ShiftOneMutator



""" Classes for defining fitness functions."""

# Author: Genevieve Hayes (Modified by Andrew Rollings)
# License: BSD 3 clause

import numpy as np


class MaxKColor:
    """Fitness function for Max-k color optimization problem. Evaluates the
    fitness of an n-dimensional state vector
    :math:`x = [x_{0}, x_{1}, \\ldots, x_{n-1}]`, where :math:`x_{i}`
    represents the color of node i, as the number of pairs of adjacent nodes
    of the same color.

    Parameters
    ----------
    edges: list of pairs
        List of all pairs of connected nodes. Order does not matter, so (a, b)
        and (b, a) are considered to be the same.

    Example
    -------
    .. highlight:: python
    .. code-block:: python

        >>> import mlrose_hiive
        >>> import numpy as np
        >>> edges = [(0, 1), (0, 2), (0, 4), (1, 3), (2, 0), (2, 3), (3, 4)]
        >>> fitness = mlrose_hiive.MaxKColor(edges)
        >>> state = np.array([0, 1, 0, 1, 1])
        >>> fitness.evaluate(state)
        3

    Note
    ----
    The MaxKColor fitness function is suitable for use in discrete-state
    optimization problems *only*.
    """

    def __init__(self, edges):

        # Remove any duplicates from list
        edges = list({tuple(sorted(edge)) for edge in edges})

        self.graph_edges = None
        self.edges = edges
        self.prob_type = 'discrete'

    def evaluate(self, state):
        """Evaluate the fitness of a state vector.

        Parameters
        ----------
        state: array
            State array for evaluation.

        Returns
        -------
        fitness: float
            Value of fitness function.
        """

        fitness = 0

        # this is the count of neigbor nodes with the same state value.
        # Therefore state value represents color.
        # This is NOT what the docs above say.

        edges = self.edges if self.graph_edges is None else self.graph_edges
        fitness = sum(int(state[n1] == state[n2]) for (n1, n2) in edges)
        """
        if fitness == 0:
            for i in range(len(edges)):
                # Check for adjacent nodes of the same color
                n1, n2 = edges[i]
                print(f'i:{i}: ({n1},{n2})[{state[n1]}] <-> [{state[n2]}]')
        """
        """
        
        if self.graph_edges is not None:
            fitness = sum(int(state[n1] == state[n2]) for (n1, n2) in self.graph_edges)
        else:
            fitness = 0
            for i in range(len(self.edges)):
                # Check for adjacent nodes of the same color
                if state[self.edges[i][0]] == state[self.edges[i][1]]:
                    fitness += 1
        """
        return fitness

    def get_prob_type(self):
        """ Return the problem type.

        Returns
        -------
        self.prob_type: string
            Specifies problem type as 'discrete', 'continuous', 'tsp'
            or 'either'.
        """
        return self.prob_type

    def set_graph(self, graph):
        self.graph_edges = [e for e in graph.edges()]



""" Crossover implementations for GA.
"""

# Author: Genevieve Hayes (modified by Andrew Rollings)
# License: BSD 3 clause

import numpy as np

from mlrose_hiive.algorithms.crossovers._crossover_base import _CrossOverBase


class OnePointCrossOver(_CrossOverBase):
    def __init__(self, opt_prob):
        super().__init__(opt_prob)

    def mate(self, p1, p2):
        n = 1 + np.random.randint(self._length-1)
        child = np.array([*p1[:n], *p2[n:]])
        return child



""" Functions to implement the randomized optimization and search algorithms.
"""

# Author: Genevieve Hayes (modified by Andrew Rollings)
# License: BSD 3 clause


def short_name(expr):
    def short_name_func_applicator(func):
        func.__short_name__ = expr
        return func
    return short_name_func_applicator


def get_short_name(v):
    return v if not hasattr(v, '__short_name__') else v.__short_name__



""" Classes for defining optimization problem objects."""

# Author: Genevieve Hayes
# License: BSD 3 clause

import numpy as np

import mlrose_hiive


class KnapsackGenerator:
    @staticmethod
    def generate(seed, number_of_items_types=10,
                 max_item_count=5, max_weight_per_item=25,
                 max_value_per_item=10, max_weight_pct=0.6,
                 multiply_by_max_item_count=True):
        np.random.seed(seed)
        weights = 1 + np.random.randint(max_weight_per_item, size=number_of_items_types)
        values = 1 + np.random.randint(max_value_per_item, size=number_of_items_types)
        problem = mlrose_hiive.KnapsackOpt(length=number_of_items_types,
                                           maximize=True, max_val=max_item_count,
                                           weights=weights, values=values,
                                           max_weight_pct=max_weight_pct,
                                           multiply_by_max_item_count=multiply_by_max_item_count)
        return problem



# Author: Genevieve Hayes (modified by Andrew Rollings)
# License: BSD 3 clause


class GeomDecay:
    """
    Schedule for geometrically decaying the simulated
    annealing temperature parameter T according to the formula:

    .. math::

        T(t) = \\max(T_{0} \\times r^{t}, T_{min})

    where:

    * :math:`T_{0}` is the initial temperature (at time t = 0);
    * :math:`r` is the rate of geometric decay; and
    * :math:`T_{min}` is the minimum temperature value.

    Parameters
    ----------
    init_temp: float, default: 1.0
        Initial value of temperature parameter T. Must be greater than 0.
    decay: float, default: 0.99
        Temperature decay parameter, r. Must be between 0 and 1.
    min_temp: float, default: 0.001
        Minimum value of temperature parameter. Must be greater than 0.

    Example
    -------
    .. highlight:: python
    .. code-block:: python

        >>> import mlrose_hiive
        >>> schedule = mlrose_hiive.GeomDecay(init_temp=10, decay=0.95, min_temp=1)
        >>> schedule.evaluate(5)
        7.73780...
    """

    def __init__(self, init_temp=1.0, decay=0.99, min_temp=0.001):

        self.init_temp = init_temp
        self.decay = decay
        self.min_temp = min_temp

        if self.init_temp <= 0:
            raise Exception("""init_temp must be greater than 0.""")

        if (self.decay <= 0) or (self.decay > 1):
            raise Exception("""decay must be between 0 and 1.""")

        if self.min_temp < 0:
            raise Exception("""min_temp must be greater than 0.""")
        elif self.min_temp > self.init_temp:
            raise Exception("""init_temp must be greater than min_temp.""")

    def evaluate(self, t):
        """Evaluate the temperature parameter at time t.

        Parameters
        ----------
        t: int
            Time at which the temperature paramter T is evaluated.

        Returns
        -------
        temp: float
            Temperature parameter at time t.
        """

        temp = self.init_temp*(self.decay**t)

        if temp < self.min_temp:
            temp = self.min_temp

        return temp

    def get_info__(self, t=None, prefix=''):
        prefix = f'_{prefix}__schedule_' if len(prefix) > 0 else 'schedule_'
        info = {
            f'{prefix}type': 'geometric',
            f'{prefix}init_temp': self.init_temp,
            f'{prefix}decay': self.decay,
            f'{prefix}min_temp': self.min_temp,
        }
        if t is not None:
            info[f'{prefix}current_value'] = self.evaluate(t)
        return info

    def __str__(self):
        return str(self.init_temp)

    def __repr__(self):
        return f'{self.__class__.__name__}(init_temp={self.init_temp}, ' \
               f'decay={self.decay}, min_temp={self.min_temp})'

    def __eq__(self, other):
        try:

            return (self.__class__.__name__ == other.__class__.__name__ and
                    self.init_temp == other.init_temp and
                    self.decay == other.decay and
                    self.min_temp == other.min_temp)
        except:
            return False


import itertools as it
from collections import defaultdict

import numpy as np

from mlrose_hiive import TSPOpt


class TSPGenerator:
    @staticmethod
    def generate(seed, number_of_cities, area_width=250, area_height=250):
        np.random.seed(seed)
        x_coords = np.random.randint(area_width, size=number_of_cities)
        y_coords = np.random.randint(area_height, size=number_of_cities)

        coords = list(tuple(zip(x_coords, y_coords)))
        duplicates = TSPGenerator.list_duplicates_(coords)

        while len(duplicates) > 0:
            for d in duplicates:
                x_coords = np.random.randint(area_width, size=len(d))
                y_coords = np.random.randint(area_height, size=len(d))
                for i in range(len(d)):
                    coords[d[i]] = (x_coords[i], y_coords[i])
                    pass
            duplicates = TSPGenerator.list_duplicates_(coords)
        distances = TSPGenerator.get_distances(coords, False)

        return TSPOpt(coords=coords, distances=distances, maximize=False)


    @staticmethod
    def get_distances(coords, truncate=True):
        distances = [(c1, c2, np.linalg.norm(np.subtract(coords[c1], coords[c2])))
                     for c1, c2 in it.product(range(len(coords)), range(len(coords)))
                     if c1 != c2 and c2 > c1]
        if truncate:
            distances = [(c1, c2, int(d)) for c1, c2, d in distances]
        return distances

    #  https://stackoverflow.com/a/5419576/40410
    @staticmethod
    def list_duplicates_(seq):
        tally = defaultdict(list)
        for i, item in enumerate(seq):
            tally[item].append(i)
        return list((indices[1:] for _, indices in tally.items() if len(indices) > 1))


if __name__=='__main__':
    results = TSPGenerator.generate(123, 22)
    print(results)



""" Classes for defining optimization problem objects."""

# Author: Genevieve Hayes (modified by Andrew Rollings)
# License: BSD 3 clause

import numpy as np

from mlrose_hiive.algorithms.crossovers import UniformCrossOver
from mlrose_hiive.algorithms.mutators import ChangeOneMutator
from mlrose_hiive.fitness import MaxKColor
from mlrose_hiive.opt_probs.discrete_opt import DiscreteOpt

import networkx as nx


class MaxKColorOpt(DiscreteOpt):
    def __init__(self, edges=None, length=None, fitness_fn=None, maximize=False,
                 max_colors=None, crossover=None, mutator=None, source_graph=None):

        if (fitness_fn is None) and (edges is None):
            raise Exception("fitness_fn or edges must be specified.")

        if length is None:
            if fitness_fn is None:
                length = len(edges)
            else:
                length = len(fitness_fn.weights)

        self.length = length

        if fitness_fn is None:
            fitness_fn = MaxKColor(edges)

        # set up initial state (everything painted one color)
        if source_graph is None:
            g = nx.Graph()
            g.add_edges_from(edges)
            self.source_graph = g
        else:
            self.source_graph = source_graph

        fitness_fn.set_graph(self.source_graph)
        # if none is provided, make a reasonable starting guess.
        # the max val is going to be the one plus the maximum number of neighbors of any one node.
        if max_colors is None:
            total_neighbor_count = [len([*self.source_graph.neighbors(n)]) for n in range(length)]
            max_colors = 1 + max(total_neighbor_count)
        self.max_val = max_colors

        crossover = UniformCrossOver(self) if crossover is None else crossover
        mutator = ChangeOneMutator(self) if mutator is None else mutator
        super().__init__(length, fitness_fn, maximize, max_colors, crossover, mutator)

        # state = [len([*g.neighbors(n)]) for n in range(length)]
        state = np.random.randint(max_colors, size=self.length)
        np.random.shuffle(state)
        # state = [0] * length
        self.set_state(state)

    def can_stop(self):
        return int(self.get_fitness()) == 0



""" Classes for defining fitness functions."""

# Author: Genevieve Hayes (Modified by Andrew Rollings)
# License: BSD 3 clause

import numpy as np

from mlrose_hiive.fitness._discrete_peaks_base import _DiscretePeaksBase


class SixPeaks(_DiscretePeaksBase):
    """Fitness function for Six Peaks optimization problem. Evaluates the
    fitness of an n-dimensional state vector :math:`x`, given parameter T, as:

    .. math::
        Fitness(x, T) = \\max(tail(0, x), head(1, x)) + R(x, T)

    where:

    * :math:`tail(b, x)` is the number of trailing b's in :math:`x`;
    * :math:`head(b, x)` is the number of leading b's in :math:`x`;
    * :math:`R(x, T) = n`, if (:math:`tail(0, x) > T` and
      :math:`head(1, x) > T`) or (:math:`tail(1, x) > T` and
      :math:`head(0, x) > T`); and
    * :math:`R(x, T) = 0`, otherwise.

    Parameters
    ----------
    t_pct: float, default: 0.1
        Threshold parameter (T) for Six Peaks fitness function, expressed as
        a percentage of the state space dimension, n (i.e.
        :math:`T = t_{pct} \\times n`).

    Example
    -------
    .. highlight:: python
    .. code-block:: python

        >>> import mlrose_hiive
        >>> import numpy as np
        >>> fitness = mlrose_hiive.SixPeaks(t_pct=0.15)
        >>> state = np.array([0, 0, 0, 1, 0, 1, 1, 0, 1, 1, 1, 1])
        >>> fitness.evaluate(state)
        12

    References
    ----------
    De Bonet, J., C. Isbell, and P. Viola (1997). MIMIC: Finding Optima by
    Estimating Probability Densities. In *Advances in Neural Information
    Processing Systems* (NIPS) 9, pp. 424–430.

    Note
    ----
    The Six Peaks fitness function is suitable for use in bit-string
    (discrete-state with :code:`max_val = 2`) optimization problems *only*.
    """

    def __init__(self, t_pct=0.1):

        self.t_pct = t_pct
        self.prob_type = 'discrete'

        if (self.t_pct < 0) or (self.t_pct > 1):
            raise Exception("""t_pct must be between 0 and 1.""")

    def evaluate(self, state):
        """Evaluate the fitness of a state vector.

        Parameters
        ----------
        state: array
            State array for evaluation.

        Returns
        -------
        fitness: float
            Value of fitness function.
        """
        _n = len(state)
        _t = np.ceil(self.t_pct*_n)

        # Calculate head and tail values
        head_0 = self.head(0, state)
        tail_0 = self.tail(0, state)
        head_1 = self.head(1, state)
        tail_1 = self.tail(1, state)

        # Calculate R(X, T)
        _r = 0
        _max_score = max(tail_0, head_1)
        if tail_0 > _t and head_1 > _t:
            _r = _n
        elif tail_1 > _t and head_0 > _t:
            _r = _n
            _max_score = max(tail_1, head_0)

        # Evaluate function
        fitness = max(tail_0, head_1) + _r

        return fitness

    def get_prob_type(self):
        """ Return the problem type.

        Returns
        -------
        self.prob_type: string
            Specifies problem type as 'discrete', 'continuous', 'tsp'
            or 'either'.
        """
        return self.prob_type


import numpy as np

from mlrose_hiive import QueensOpt


class QueensGenerator:
    @staticmethod
    def generate(seed, size=20):
        np.random.seed(seed)
        problem = QueensOpt(length=size)
        return problem



""" Classes for defining fitness functions."""

# Author: Genevieve Hayes (Modified by Andrew Rollings)
# License: BSD 3 clause

import numpy as np


class FlipFlop:
    """Fitness function for Flip Flop optimization problem. Evaluates the
    fitness of a state vector :math:`x` as the total number of pairs of
    consecutive elements of :math:`x`, (:math:`x_{i}` and :math:`x_{i+1}`)
    where :math:`x_{i} \\neq x_{i+1}`.

    Example
    -------
    .. highlight:: python
    .. code-block:: python

        >>> import mlrose_hiive
        >>> import numpy as np
        >>> fitness = mlrose_hiive.FlipFlop()
        >>> state = np.array([0, 1, 0, 1, 1, 1, 1])
        >>> fitness.evaluate(state)
        3

    Note
    ----
    The Flip Flop fitness function is suitable for use in discrete-state
    optimization problems *only*.
    """

    def __init__(self):

        self.prob_type = 'discrete'

    def evaluate(self, state):
        """Evaluate the fitness of a state vector.

        Parameters
        ----------
        state: array
            State array for evaluation.

        Returns
        -------
        fitness: float
            Value of fitness function.
        """

        fitness = sum([state[i] != state[i-1] for i in range(1, len(state))])


        # may not be faster
        """
        runs = np.zeros(state.size-1, dtype=int)
        np.not_equal(state[:-1], state[1:], out=runs)
        fitness = np.sum(runs)
        """

        return fitness

    def evaluate_many(self, states):
        """Evaluate the fitness of a an ndarray of states.

        Parameters
        ----------
        states: ndarray
            States array for evaluation.

        Returns
        -------
        fitness: ndarray
            Population fitness values.
        """

        runs = np.zeros((states.shape[0], states.shape[1]-1), dtype=int)
        np.not_equal(states[:, :-1], states[:, 1:], out=runs)
        fitness = np.sum(runs, axis=1)

        return fitness

    def get_prob_type(self):
        """ Return the problem type.

        Returns
        -------
        self.prob_type: string
            Specifies problem type as 'discrete', 'continuous', 'tsp'
            or 'either'.
        """
        return self.prob_type



import mlrose_hiive
from mlrose_hiive.decorators import short_name
from mlrose_hiive.runners._runner_base import _RunnerBase

"""
Example usage:

    experiment_name = 'example_experiment'
    problem = TSPGenerator.generate(seed=SEED, number_of_cities=22)

    ga = GARunner(problem=problem,
                  experiment_name=experiment_name,
                  output_directory=OUTPUT_DIRECTORY,
                  seed=SEED,
                  iteration_list=2 ** np.arange(12),
                  max_attempts=1000,
                  population_sizes=[150, 200, 300],
                  mutation_rates=[0.4, 0.5, 0.6])
                  
    # the two data frames will contain the results
    df_run_stats, df_run_curves = ga.run()                  
"""


@short_name('ga')
class GARunner(_RunnerBase):

    def __init__(self, problem, experiment_name, seed, iteration_list, population_sizes, mutation_rates,
                 hamming_factors=None, hamming_factor_decays=None, max_attempts=500, generate_curves=True, **kwargs):
        super().__init__(problem=problem, experiment_name=experiment_name, seed=seed, iteration_list=iteration_list,
                         max_attempts=max_attempts, generate_curves=generate_curves,
                         **kwargs)
        self.population_sizes = population_sizes
        self.mutation_rates = mutation_rates
        self.hamming_factors = hamming_factors
        self.hamming_factor_decays = hamming_factor_decays

    def run(self):
        return super().run_experiment_(algorithm=mlrose_hiive.genetic_alg,
                                       pop_size=('Population Size', self.population_sizes),
                                       mutation_prob=('Mutation Rate', self.mutation_rates),
                                       hamming_factor=('Hamming Factor', self.hamming_factors),
                                       hamming_decay_factor=('Hamming Factor Decay Rate', self.hamming_factor_decays))







""" GA Mutators."""

# Author: Genevieve Hayes
# Modified: Andrew Rollings
# License: BSD 3 clause
import numpy as np

from mlrose_hiive.algorithms.mutators._mutator_base import _MutatorBase


class ShiftOneMutator(_MutatorBase):

    def __init__(self, opt_prob):
        super().__init__(opt_prob)
        self._max_val = opt_prob.max_val

    def mutate(self, child, mutation_probability):
        if np.random.rand() < mutation_probability:
            # do shift one mutation
            m = np.random.randint(len(child))
            # bump value up or down
            new_val = child[m] + (1 if np.random.randint(2) == 0 else -1)
            # wrap around if greater than or less than max_val
            new_val = (self._max_val + new_val) % self._max_val
            child[m] = new_val
        return child



""" Classes for defining decay schedules for simulated annealing."""

# Author: Genevieve Hayes
# License: BSD 3 clause


class CustomSchedule:
    """Class for generating your own temperature schedule.

    Parameters
    ----------
    schedule: callable
        Function for calculating the temperature at time t with the signature
        :code:`schedule(t, **kwargs)`.

    kwargs: additional arguments
        Additional parameters to be passed to schedule.

    Example
    -------
    .. highlight:: python
    .. code-block:: python

        >>> import mlrose_hiive
        >>> def custom(t, c): return t + c
        >>> kwargs = {'c': 10}
        >>> schedule = mlrose_hiive.CustomSchedule(custom, **kwargs)
        >>> schedule.evaluate(5)
        15
    """

    def __init__(self, schedule, **kwargs):

        self.schedule = schedule
        self.kwargs = kwargs

    def evaluate(self, t):
        """Evaluate the temperature parameter at time t.

        Parameters
        ----------
        t: int
            Time at which the temperature paramter T is evaluated.

        Returns
        -------
        temp: float
            Temperature parameter at time t.
        """

        temp = self.schedule(t, **self.kwargs)
        return temp

    def get_info__(self, t=None, prefix=''):
        prefix = f'_{prefix}__schedule_' if len(prefix) > 0 else 'schedule_'
        info = {
            f'{prefix}type': 'custom',
            f'{prefix}schedule': self.schedule
        }
        info.update({f'{prefix}_args_{k}': v for k, v in self.kwargs.items()})
        if t is not None:
            info[f'{prefix}current_value'] = self.evaluate(t)
        return info

    def __str__(self):
        return str(self.schedule)

    def __repr__(self):
        return f'{self.__class__.__name__}[{self.__dict__}]'

    def __eq__(self, other):
        try:

            return (self.__class__.__name__ == other.__class__.__name__ and
                    self.get_info__() == other.get_info__())
        except:
            return False



""" Classes for defining neural network weight optimization problems."""

# Author: Genevieve Hayes (modified by Andrew Rollings)
# License: BSD 3 clause


from sklearn.base import ClassifierMixin

from mlrose_hiive.algorithms.decay import GeomDecay
from ._nn_core import _NNCore


class LogisticRegression(_NNCore, ClassifierMixin):
    """Class for defining logistic regression weights optimization
    problem. Inherits :code:`fit` and :code:`predict` methods from
    :code:`NeuralNetwork()` class.

    Parameters
    ----------
    algorithm: string, default: 'random_hill_climb'
        Algorithm used to find optimal network weights. Must be one
        of:'random_hill_climb', 'simulated_annealing', 'genetic_alg' or
        'gradient_descent'.

    max_iters: int, default: 100
        Maximum number of iterations used to fit the weights.

    bias: bool, default: True
        Whether to include a bias term.

    learning_rate: float, default: 0.1
        Learning rate for gradient descent or step size for randomized
        optimization algorithms.

    early_stopping: bool, default: False
        Whether to terminate algorithm early if the loss is not improving.
        If :code:`True`, then stop after max_attempts iters with no
        improvement.

    clip_max: float, default: 1e+10
        Used to limit weights to the range [-1*clip_max, clip_max].

    restarts: int, default: 0
        Number of random restarts.
        Only required if :code:`algorithm = 'random_hill_climb'`.

    schedule: schedule object, default = mlrose_hiive.GeomDecay()
        Schedule used to determine the value of the temperature parameter.
        Only required if :code:`algorithm = 'simulated_annealing'`.

    pop_size: int, default: 200
        Size of population. Only required if :code:`algorithm = 'genetic_alg'`.

    mutation_prob: float, default: 0.1
        Probability of a mutation at each element of the state vector during
        reproduction, expressed as a value between 0 and 1. Only required if
        :code:`algorithm = 'genetic_alg'`.

    max_attempts: int, default: 10
        Maximum number of attempts to find a better state. Only required if
        :code:`early_stopping = True`.

    random_state: int, default: None
        If random_state is a positive integer, random_state is the seed used
        by np.random.seed(); otherwise, the random seed is not set.

    curve: bool, default: False
        If bool is true, curve containing the fitness at each training
        iteration is returned.

    Attributes
    ----------
    fitted_weights: array
        Numpy array giving the fitted weights when :code:`fit` is performed.

    loss: float
        Value of loss function for fitted weights when :code:`fit` is
        performed.

    fitness_curve: array
        Numpy array giving the fitness at each training iteration.
    """

    def __init__(self, algorithm='random_hill_climb', max_iters=100, bias=True,
                 learning_rate=0.1, early_stopping=False, clip_max=1e+10,
                 restarts=0, schedule=GeomDecay(), pop_size=200,
                 mutation_prob=0.1, max_attempts=10, random_state=None,
                 curve=False):

        _NNCore.__init__(
            self, hidden_nodes=[], activation='sigmoid',
            algorithm=algorithm, max_iters=max_iters, bias=bias,
            is_classifier=True, learning_rate=learning_rate,
            early_stopping=early_stopping, clip_max=clip_max,
            restarts=restarts, schedule=schedule, pop_size=pop_size,
            mutation_prob=mutation_prob, max_attempts=max_attempts,
            random_state=random_state, curve=curve)



import mlrose_hiive.neural.activation as act
import sklearn.metrics as skmt

from sklearn.neural_network import MLPClassifier
from sklearn.base import BaseEstimator
from mlrose_hiive.decorators import short_name
from mlrose_hiive.runners._nn_runner_base import _NNRunnerBase


@short_name('skmlp')
class SKMLPRunner(_NNRunnerBase):
    class _MLPClassifier(BaseEstimator):
        def __init__(self, runner, **kwargs):
            self.runner = runner
            self.mlp = MLPClassifier(**kwargs)
            self.state_callback = self.runner._save_state
            self.fit_started_ = False
            self.user_info_ = None
            self.kwargs_ = kwargs

            self.loss_ = 1.
            self.state_ = None
            self.curve_ = []

            # need to intercept the classifier so we can track statistics.
            if runner.generate_curves:
                if hasattr(self.mlp, '_update_no_improvement_count'):
                    self._mlp_update_no_improvement_count = self.mlp._update_no_improvement_count
                    self.mlp._update_no_improvement_count = self._update_no_improvement_count_intercept
                if hasattr(self.mlp, '_loss_grad_lbfgs'):
                    self._mlp_loss_grad_lbfgs = self.mlp._loss_grad_lbfgs
                    self.mlp._loss_grad_lbfgs = self._loss_grad_lbfgs_intercept

        def __getattr__(self, item, default=None):
            if 'mlp' in self.__dict__ and hasattr(self.__dict__['mlp'], item):
                return self.__dict__['mlp'].__getattr__(item, default)
            return self.__dict__[item] if item in self.__dict__ else default

        def __setattr__(self, item, value):
            if 'mlp' in self.__dict__ and hasattr(self.__dict__['mlp'], item):
                self.__dict__['mlp'].__setattr__(item, value)
            self.__dict__[item] = value

        def get_params(self, deep=True):
            out = super().get_params()
            out.update(self.mlp.get_params())
            # exclude any that end with an underscore
            out = {k: v for (k, v) in out.items() if not k[-1] == '_'}
            return out

        def fit(self, x_train, y_train=None):
            self.fit_started_ = True
            self.runner._start_run_timing()
            # make initial callback
            self._invoke_runner_callback()
            return self.mlp.fit(x_train, y_train)

        def predict(self, x_test):
            return self.mlp.predict(x_test)

        def _update_no_improvement_count_intercept(self, early_stopping, x_val, y_val):
            ret = self._mlp_update_no_improvement_count(early_stopping, x_val, y_val)
            self._state = self.mlp.coefs_ if hasattr(self.mlp, 'coefs_') else []
            self.loss_ = self.mlp.loss_ if hasattr(self.mlp, 'loss_') else 0
            if hasattr(self.mlp, 'loss_curve_'):
                self.curve_ = self.mlp.loss_curve_
            else:
                self.curve_.append(self.loss_)
            self._invoke_runner_callback()
            return ret

        def _loss_grad_lbfgs_intercept(self, packed_coef_inter, x, y, activations, deltas, coef_grads, intercept_grads):
            f, g = self._mlp_loss_grad_lbfgs(packed_coef_inter, x, y, activations, deltas,
                                             coef_grads, intercept_grads)
            self.loss_ = f
            self.state_ = g
            self.curve_.append(self.loss_)
            self._invoke_runner_callback()
            return f, g

        def _invoke_runner_callback(self):
            iterations = self.mlp.n_iter_ if hasattr(self.mlp, 'n_iter_') else 0
            no_improvement_count = self.mlp._no_improvement_count if hasattr(self.mlp, '_no_improvement_count') else 0

            done = (self.mlp.early_stopping and (no_improvement_count > self.mlp.n_iter_no_change) or
                    iterations == self.mlp.max_iter)

            # check for early abort.
            if self.runner.has_aborted():
                return self
            if self.user_info_ is None:
                self.user_info_ = [(k, self.__dict__[k]) for k in self.kwargs_.keys() if hasattr(self, k)]
                for k, v in self.user_info_:
                    self.runner._log_current_argument(k, v)

            return self.state_callback(iteration=iterations,
                                       state=self.state_,
                                       fitness=self.loss_,
                                       user_data=self.user_info_,
                                       attempt=no_improvement_count,
                                       done=done,
                                       curve=self.curve_)

    def __init__(self, x_train, y_train, x_test, y_test, experiment_name, seed, iteration_list,
                 grid_search_parameters, grid_search_scorer_method=skmt.balanced_accuracy_score,
                 early_stopping=True, max_attempts=500, n_jobs=1, cv=5,
                 generate_curves=True, output_directory=None, replay=False, **kwargs):

        # take a copy of the grid search parameters
        grid_search_parameters = {**grid_search_parameters}

        # hack for compatibility purposes
        if 'max_iters' in grid_search_parameters:
            grid_search_parameters['max_iter'] = grid_search_parameters.pop('max_iters')

        if 'max_attempts' in grid_search_parameters:
            grid_search_parameters['n_iter_no_change'] = grid_search_parameters.pop('max_attempts')

        super().__init__(x_train=x_train, y_train=y_train, x_test=x_test, y_test=y_test,
                         experiment_name=experiment_name,
                         seed=seed,
                         iteration_list=iteration_list,
                         grid_search_parameters=grid_search_parameters,
                         grid_search_scorer_method=grid_search_scorer_method,
                         generate_curves=generate_curves,
                         output_directory=output_directory,
                         replay=replay,
                         n_jobs=n_jobs,
                         cv=cv)

        # build the classifier
        self.classifier = self._MLPClassifier(runner=self,
                                              shuffle=True,
                                              random_state=seed,
                                              verbose=False,
                                              warm_start=False,
                                              early_stopping=early_stopping,
                                              n_iter_no_change=max_attempts,
                                              **kwargs)

        self.classifier.runner = self

    @staticmethod
    def build_grid_search_parameters(grid_search_parameters, **kwargs):
        # extract nn parameters
        all_grid_search_parameters = _NNRunnerBase.build_grid_search_parameters(grid_search_parameters, **kwargs)
        # make sure activation set is the right type.
        if 'activation' in all_grid_search_parameters:
            activation_set = list(all_grid_search_parameters['activation'])
            for i in range(len(activation_set)):
                a = activation_set[i]
                if a == act.relu:
                    activation_set[i] = 'relu'
                elif a == act.sigmoid:
                    activation_set[i] = 'logistic'
                elif a == act.tanh:
                    activation_set[i] = 'tanh'
                elif a == act.identity:
                    activation_set[i] = 'identity'
                elif a == act.softmax:
                    activation_set[i] = 'softmax'
            all_grid_search_parameters['activation'] = activation_set
        return all_grid_search_parameters



""" Classes for defining optimization problem objects."""

# Author: Andrew Rollings
# License: BSD 3 clause

import numpy as np

from mlrose_hiive import DiscreteOpt, ContinuousPeaks


class ContinuousPeaksGenerator:
    @staticmethod
    def generate(seed, size=20, t_pct=0.1):
        np.random.seed(seed)
        fitness = ContinuousPeaks(t_pct=t_pct)
        problem = DiscreteOpt(length=size, fitness_fn=fitness)
        return problem



""" Classes for defining fitness functions."""

# Author: Genevieve Hayes (Modified by Andrew Rollings)
# License: BSD 3 clause

import numpy as np


class Knapsack:
    """Fitness function for Knapsack optimization problem. Given a set of n
    items, where item i has known weight :math:`w_{i}` and known value
    :math:`v_{i}`; and maximum knapsack capacity, :math:`W`, the Knapsack
    fitness function evaluates the fitness of a state vector
    :math:`x = [x_{0}, x_{1}, \\ldots, x_{n-1}]` as:

    .. math::
        Fitness(x) = \\sum_{i = 0}^{n-1}v_{i}x_{i}, \\text{ if}
        \\sum_{i = 0}^{n-1}w_{i}x_{i} \\leq W, \\text{ and 0, otherwise,}

    where :math:`x_{i}` denotes the number of copies of item i included in the
    knapsack.

    Parameters
    ----------
    weights: list
        List of weights for each of the n items.

    values: list
        List of values for each of the n items.

    max_weight_pct: float, default: 0.35
        Parameter used to set maximum capacity of knapsack (W) as a percentage
        of the total of the weights list
        (:math:`W =` max_weight_pct :math:`\\times` total_weight).

    Example
    -------
    .. highlight:: python
    .. code-block:: python

        >>> import mlrose_hiive
        >>> import numpy as np
        >>> weights = [10, 5, 2, 8, 15]
        >>> values = [1, 2, 3, 4, 5]
        >>> max_weight_pct = 0.6
        >>> fitness = mlrose_hiive.Knapsack(weights, values, max_weight_pct)
        >>> state = np.array([1, 0, 2, 1, 0])
        >>> fitness.evaluate(state)
        11

    Note
    ----
    The Knapsack fitness function is suitable for use in discrete-state
    optimization problems *only*.
    """

    def __init__(self, weights, values, max_weight_pct=0.35, max_item_count=1, multiply_by_max_item_count=False):

        self.weights = weights
        self.values = values
        count_multiplier = max_item_count if multiply_by_max_item_count else 1.0
        self._w = np.ceil(np.sum(self.weights) * max_weight_pct * count_multiplier)
        self.prob_type = 'discrete'

        if len(self.weights) != len(self.values):
            raise Exception("""The weights array and values array must be"""
                            + """ the same size.""")

        if min(self.weights) <= 0:
            raise Exception("""All weights must be greater than 0.""")

        if min(self.values) <= 0:
            raise Exception("""All values must be greater than 0.""")

        if max_item_count <= 0:
            raise Exception("""max_item_count must be greater than 0.""")

        if max_weight_pct <= 0:
            raise Exception("""max_weight_pct must be greater than 0.""")

    def evaluate(self, state):
        """Evaluate the fitness of a state vector.

        Parameters
        ----------
        state: array
            State array for evaluation. Must be the same length as the weights
            and values arrays.

        Returns
        -------
        fitness: float
            Value of fitness function.
        """

        if len(state) != len(self.weights):
            raise Exception("""The state array must be the same size as the"""
                            + """ weight and values arrays.""")

        # Calculate total weight and value of knapsack
        total_weight = np.sum(state*self.weights)
        total_value = np.sum(state*self.values)

        # Allow for weight constraint
        if total_weight <= self._w:
            fitness = total_value
        else:
            fitness = 0

        return fitness

    def get_prob_type(self):
        """ Return the problem type.

        Returns
        -------
        self.prob_type: string
            Specifies problem type as 'discrete', 'continuous', 'tsp'
            or 'either'.
        """
        return self.prob_type



""" Neural network activation functions."""

# Author: Genevieve Hayes (Modified by Andrew Rollings)
# License: BSD 3 clause
from mlrose_hiive.decorators import short_name

import numpy as np

import warnings
warnings.filterwarnings("ignore")


@short_name('sigmoid')
def sigmoid(x, deriv=False):
    """Sigmoid activation function

    Parameters
    ----------
    x: array
        Array containing input data.

    deriv: bool, default: False
        Whether to return the function or its derivative.
        Set True for derivative.

    Returns
    -------
    fx: array
        Value of activation function at x
    """
    fx = 1/(1 + np.exp(-x))

    if deriv:
        fx *= (1 - fx)

    return fx



import mlrose_hiive
from mlrose_hiive.decorators import short_name

from mlrose_hiive.runners._runner_base import _RunnerBase

"""
Example usage:

    experiment_name = 'example_experiment'
    problem = TSPGenerator.generate(seed=SEED, number_of_cities=22)

    rhc = RHCRunner(problem=problem,
                    experiment_name=experiment_name,
                    output_directory=OUTPUT_DIRECTORY,
                    seed=SEED,
                    iteration_list=2 ** np.arange(10),
                    max_attempts=5000,
                    restart_list=[25, 75, 100])   

    # the two data frames will contain the results
    df_run_stats, df_run_curves = rhc.run()               
"""


@short_name('rhc')
class RHCRunner(_RunnerBase):

    def __init__(self, problem, experiment_name, seed, iteration_list, restart_list,
                 max_attempts=500, generate_curves=True, **kwargs):
        super().__init__(problem=problem, experiment_name=experiment_name, seed=seed, iteration_list=iteration_list,
                         max_attempts=max_attempts, generate_curves=generate_curves,
                         **kwargs)
        self.restart_list = restart_list

    def run(self):
        return super().run_experiment_(algorithm=mlrose_hiive.random_hill_climb,
                                       restarts=('Restarts', self.restart_list))



""" Classes for defining algorithms problem objects."""

# Author: Genevieve Hayes (modified by Andrew Rollings)
# License: BSD 3 clause

from .ga import (genetic_alg)
from .sa import (simulated_annealing)
from .hc import (hill_climb)
from .rhc import (random_hill_climb)
from .gd import (gradient_descent)
from .mimic import (mimic)

from .crossovers import UniformCrossOver, TSPCrossOver, OnePointCrossOver

from .decay import ArithDecay, CustomSchedule, ExpDecay, GeomDecay
from .mutators import ChangeOneMutator, DiscreteMutator, ShiftOneMutator, SwapMutator



""" Classes for defining neural network weight optimization problems."""

# Author: Genevieve Hayes (modified by Andrew Rollings)
# License: BSD 3 clause

from .weights import (flatten_weights, unflatten_weights, gradient_descent_original)



""" Classes for defining optimization problem objects."""

# Author: Genevieve Hayes
# License: BSD 3 clause

from .max_k_color_generator import MaxKColorGenerator
from .knapsack_generator import KnapsackGenerator
from .flip_flop_generator import FlipFlopGenerator
from .queens_generator import QueensGenerator
from .tsp_generator import TSPGenerator
from .continuous_peaks_generator import ContinuousPeaksGenerator



""" Classes for defining neural network weight optimization problems."""

# Author: Genevieve Hayes (modified by Andrew Rollings)
# License: BSD 3 clause

import numpy as np

from mlrose_hiive.neural._nn_base import _NNBase


class NNClassifier(_NNBase):

    def __init__(self,
                 runner,
                 algorithm=None,
                 # grid-search
                 activation=None,
                 hidden_layer_sizes=None,
                 max_iters=100,
                 max_attempts=10,
                 learning_rate_init=0.1,
                 bias=True,
                 early_stopping=False,
                 clip_max=1e+10,
                 seed=None,
                 **kwargs
                 ):
        super().__init__()

        self.runner = runner
        self.grid_search_parameters = runner.grid_search_parameters

        # nn specific properties
        #  (grid-search settable)
        self.hidden_layer_sizes = hidden_layer_sizes
        self.activation = activation
        self.learning_rate_init = learning_rate_init
        self.max_iters = max_iters
        self.max_attempts = max_attempts

        #  (non-grid-search settable)
        self.bias = bias
        self.early_stopping = early_stopping
        self.clip_max = clip_max
        self.algorithm = algorithm

        # result properties
        self.fitness_fn = None
        self.problem = None
        self.fitted_weights = None
        self.output_activation = None
        self.predicted_probabilities = None
        self.node_list = None
        self.node_count = None
        self.loss = None
        self.fit_started_ = False
        self.seed = seed

        # extra parameters
        self.kwargs = kwargs
        for k, v in kwargs.items():
            if hasattr(self, k):
                continue
            self.__setattr__(k, v)

    def get_params(self, deep=True):
        out = super().get_params(deep)
        # exclude any that end with an underscore
        out = {k: v for (k, v) in out.items() if not k[-1] == '_'}
        ap = {k: self.__dict__[k] if k in self.__dict__ else None for k in self.grid_search_parameters if k not in out}
        out.update(ap)
        return out

    def _get_nodes(self, x_train, y_train):
        return _NNBase._build_node_list(X=x_train, y=y_train,
                                        hidden_nodes=self.hidden_layer_sizes,
                                        bias=self.bias)

    def fit(self, x_train, y_train=None, init_weights=None):

        x_train, y_train = self._format_x_y_data(x_train, y_train)
        self.node_list = self._get_nodes(x_train, y_train)
        self.node_count = _NNBase._calculate_state_size(self.node_list)
        fitness, problem = _NNBase._build_problem_and_fitness_function(X=x_train,
                                                                       y=y_train,
                                                                       node_list=self.node_list,
                                                                       activation=self.activation,
                                                                       learning_rate=self.learning_rate_init,
                                                                       bias=self.bias,
                                                                       clip_max=self.clip_max)
        self.fitness_fn = fitness
        self.problem = problem

        # check for early abort.
        if self.runner.has_aborted() or self.runner.replay_mode():
            self.fitted_weights = np.array([np.NaN] * self.node_count)
            self.loss = np.NaN
            self.output_activation = self.fitness_fn.get_output_activation()
            return self

        if self.algorithm is not None:
            # self._perform_grid_search()
            params = {k: self.__getattribute__(k) for k in self.kwargs}
            if init_weights is None:
                np.random.seed(self.seed)
                init_weights = np.random.uniform(-1, 1, self.node_count)
            params['init_state'] = init_weights
            total_args = {
                'algorithm': self.algorithm,
                'activation': self.activation,
                'bias': self.bias,
                'early_stopping': self.early_stopping,
                'clip_max': self.clip_max,
                'hidden_layer_sizes': self.hidden_layer_sizes,
                'learning_rate_init': self.learning_rate_init
            }
            max_attempts = self.max_attempts if self.early_stopping else self.max_iters
            self.fit_started_ = True
            fitted_weights, loss, _ = self.runner.run_one_experiment_(algorithm=self.algorithm,
                                                                      problem=problem,
                                                                      max_iters=self.max_iters,
                                                                      max_attempts=max_attempts,
                                                                      total_args=total_args,
                                                                      **params)

            # Save fitted weights
            self.fitted_weights = problem.get_state()
            self.loss = loss
            self.output_activation = self.fitness_fn.get_output_activation()

        return self

    def predict(self, x_test):
        if not np.shape(x_test)[1] == (self.node_list[0] - self.bias):
            raise Exception("""The number of columns in X must equal %d"""
                            % ((self.node_list[0] - self.bias),))

        y_pred, pp = self._predict(X=x_test,
                                   fitted_weights=self.fitted_weights,
                                   node_list=self.node_list,
                                   input_activation=self.activation,
                                   output_activation=self.output_activation,
                                   bias=self.bias)

        self.predicted_probabilities = pp
        return y_pred

    def predict_proba(self, x_test):
        self.predict(x_test)
        return self.predicted_probabilities



""" Classes for defining neural network weight optimization problems."""

# Author: Genevieve Hayes (modified by Andrew Rollings)
# License: BSD 3 clause

from .activation import (identity, relu, sigmoid, softmax, tanh)
from .utils import (flatten_weights, unflatten_weights)
from .neural_network import NeuralNetwork
from .fitness.network_weights import NetworkWeights
from .logistic_regression import LogisticRegression
from .linear_regression import LinearRegression
from .nn_classifier import NNClassifier



""" Classes for defining fitness functions."""

# Author: Genevieve Hayes (Modified by Andrew Rollings)
# License: BSD 3 clause

import numpy as np
import pandas as pd


class TravellingSales:
    """Fitness function for Travelling Salesman optimization problem.
    Evaluates the fitness of a tour of n nodes, represented by state vector
    :math:`x`, giving the order in which the nodes are visited, as the total
    distance travelled on the tour (including the distance travelled between
    the final node in the state vector and the first node in the state vector
    during the return leg of the tour). Each node must be visited exactly
    once for a tour to be considered valid.

    Parameters
    ----------
    coords: list of pairs, default: None
        Ordered list of the (x, y) coordinates of all nodes (where element i
        gives the coordinates of node i). This assumes that travel between
        all pairs of nodes is possible. If this is not the case, then use
        :code:`distances` instead.

    distances: list of triples, default: None
        List giving the distances, d, between all pairs of nodes, u and v, for
        which travel is possible, with each list item in the form (u, v, d).
        Order of the nodes does not matter, so (u, v, d) and (v, u, d) are
        considered to be the same. If a pair is missing from the list, it is
        assumed that travel between the two nodes is not possible. This
        argument is ignored if coords is not :code:`None`.

    Examples
    --------
    .. highlight:: python
    .. code-block:: python

        >>> import mlrose_hiive
        >>> import numpy as np
        >>> coords = [(0, 0), (3, 0), (3, 2), (2, 4), (1, 3)]
        >>> dists = [(0, 1, 3), (0, 2, 5), (0, 3, 1), (0, 4, 7), (1, 3, 6),
                     (4, 1, 9), (2, 3, 8), (2, 4, 2), (3, 2, 8), (3, 4, 4)]
        >>> fitness_coords = mlrose_hiive.TravellingSales(coords=coords)
        >>> state = np.array([0, 1, 4, 3, 2])
        >>> fitness_coords.evaluate(state)
        13.86138...
        >>> fitness_dists = mlrose_hiive.TravellingSales(distances=dists)
        >>> fitness_dists.evaluate(state)
        29

    Note
    ----
    1. The TravellingSales fitness function is suitable for use in travelling
       salesperson (tsp) optimization problems *only*.
    2. It is necessary to specify at least one of :code:`coords` and
       :code:`distances` in initializing a TravellingSales fitness function
       object.
    """

    def __init__(self, coords=None, distances=None):

        if coords is None and distances is None:
            raise Exception("""At least one of coords and distances must be"""
                            + """ specified.""")

        elif coords is not None:
            self.is_coords = True
            path_list = []
            dist_list = []

        else:
            self.is_coords = False

            # Remove any duplicates from list
            distances = list({tuple(sorted(dist[0:2]) + [dist[2]])
                              for dist in distances})

            # Split into separate lists
            node1_list, node2_list, dist_list = zip(*distances)

            if min(dist_list) <= 0:
                raise Exception("""The distance between each pair of nodes"""
                                + """ must be greater than 0.""")
            if min(node1_list + node2_list) < 0:
                raise Exception("""The minimum node value must be 0.""")

            if not max(node1_list + node2_list) == \
                    (len(set(node1_list + node2_list)) - 1):
                raise Exception("""All nodes must appear at least once in"""
                                + """ distances.""")

            path_list = list(zip(node1_list, node2_list))

        self.coords = coords
        self.distances = distances
        self.path_list = path_list
        self.dist_list = dist_list
        self.prob_type = 'tsp'
        if self.coords:
            self.calculate_fitness = self.__calculate_fitness_by_coords
        else:
            self.df_path_list = pd.DataFrame([[self.path_list[i][0], self.path_list[i][1], self.dist_list[i]] for i in range(len(self.path_list))])
            self.calculate_fitness = self.__calculate_fitness_by_distance

    def evaluate(self, state):
        """Evaluate the fitness of a state vector.

        Parameters
        ----------
        state: array
            State array for evaluation. Each integer between 0 and
            (len(state) - 1), inclusive must appear exactly once in the array.

        Returns
        -------
        fitness: float
            Value of fitness function. Returns :code:`np.inf` if travel between
            two consecutive nodes on the tour is not possible.
        """

        if self.is_coords and len(state) != len(self.coords):
            raise Exception("""state must have the same length as coords.""")

        if not len(state) == len(set(state)):
            raise Exception("""Each node must appear exactly once in state.""")

        if min(state) < 0:
            raise Exception("""All elements of state must be non-negative"""
                            + """ integers.""")

        if max(state) >= len(state):
            raise Exception("""All elements of state must be less than"""
                            + """ len(state).""")

        return self.calculate_fitness(state)

    def __calculate_fitness_by_coords(self, state):
        # Calculate length of journey
        ls = len(state)
        nodes = np.array([self.coords[state[i]] for i in range(ls)] + [self.coords[state[0]]])
        nodes.reshape((2, nodes.size // 2))
        fitness = np.linalg.norm(nodes[1:] - nodes[:-1], axis=1).sum()

        return fitness

    def __calculate_fitness_by_distance(self, state):

        ls = len(state)

        nodes = np.array([[state[i-1], state[i]] for i in range(1, ls)] + [[state[ls-1]] + [state[0]]])
        nodes.sort(axis=1)
        df_nodes = pd.merge(self.df_path_list, pd.DataFrame(nodes), how='inner')
        if df_nodes.shape[0] != nodes.shape[0]:
            return np.inf
        fitness = df_nodes.iloc[:, 2].sum()
        return fitness

    def get_prob_type(self):
        """ Return the problem type.

        Returns
        -------
        self.prob_type: string
            Specifies problem type as 'discrete', 'continuous', 'tsp'
            or 'either'.
        """
        return self.prob_type



""" Classes for defining fitness functions."""

# Author: Genevieve Hayes (Modified by Andrew Rollings)
# License: BSD 3 clause

import numpy as np


class OneMax:
    """Fitness function for One Max optimization problem. Evaluates the
    fitness of an n-dimensional state vector
    :math:`x = [x_{0}, x_{1}, \\ldots, x_{n-1}]` as:

    .. math::
        Fitness(x) = \\sum_{i = 0}^{n-1}x_{i}

    Example
    -------
    .. highlight:: python
    .. code-block:: python

        >>> import mlrose_hiive
        >>> import numpy as np
        >>> fitness = mlrose_hiive.OneMax()
        >>> state = np.array([0, 1, 0, 1, 1, 1, 1])
        >>> fitness.evaluate(state)
        5

    Note
    -----
    The One Max fitness function is suitable for use in either discrete or
    continuous-state optimization problems.
    """

    def __init__(self):

        self.prob_type = 'either'

    def evaluate(self, state):
        """Evaluate the fitness of a state vector.

        Parameters
        ----------
        state: array
            State array for evaluation.

        Returns
        -------
        fitness: float
            Value of fitness function.
        """

        fitness = np.sum(state)
        return fitness

    def get_prob_type(self):
        """ Return the problem type.

        Returns
        -------
        self.prob_type: string
            Specifies problem type as 'discrete', 'continuous', 'tsp'
            or 'either'.
        """
        return self.prob_type



""" Classes for defining fitness functions."""

# Author: Genevieve Hayes (Modified by Andrew Rollings)
# License: BSD 3 clause


class _DiscretePeaksBase:

    @staticmethod
    def head(_b, _x):
        """Determine the number of leading b's in vector x.

        Parameters
        ----------
        b: int
            Integer for counting at head of vector.
        x: array
            Vector of integers.

        Returns
        -------
        head: int
            Number of leading b's in x.
        """

        # Initialize counter
        _head = 0

        # Iterate through values in vector
        for i in _x:
            if i == _b:
                _head += 1
            else:
                break

        return _head

    @staticmethod
    def tail(_b, _x):
        """Determine the number of trailing b's in vector x.

        Parameters
        ----------
        b: int
            Integer for counting at tail of vector.

        x: array
            Vector of integers.

        Returns
        -------
        tail: int
            Number of trailing b's in x.
        """

        # Initialize counter
        _tail = 0

        # Iterate backwards through values in vector
        for i in range(len(_x)):
            if _x[len(_x) - i - 1] == _b:
                _tail += 1
            else:
                break

        return _tail




import os


def build_data_filename(output_directory, runner_name, experiment_name, df_name, x_param='', y_param='', ext=''):
    # ensure directory exists
    try:
        os.makedirs(os.path.join(output_directory, experiment_name))
    except:
        pass

    # return full filename
    if len(ext) > 0 and not ext[0] == '.':
        ext = f'.{ext}'
    if len(x_param) > 0 and not x_param[0] == '_' and not x_param[-1] == '_':
        x_param = f'_{x_param}_'
    if len(y_param) > 0 and not y_param[0] == '_' and not y_param[-1] == '_':
        y_param = f'_{y_param}'
    return os.path.join(output_directory,
                        experiment_name,
                        f'{runner_name.lower()}__{experiment_name}__{df_name}{x_param}{y_param}{ext}')


""" GA Crossovers."""

# Author: Genevieve Hayes
# License: BSD 3 clause

from .uniform_crossover import UniformCrossOver
from .tsp_crossover import TSPCrossOver
from .one_point_crossover import OnePointCrossOver



""" Classes for defining neural network weight optimization problems."""

# Author: Genevieve Hayes (modified by Andrew Rollings)
# License: BSD 3 clause


import numpy as np
from abc import abstractmethod
from mlrose_hiive.algorithms.decay import GeomDecay
from mlrose_hiive.algorithms.rhc import random_hill_climb
from mlrose_hiive.algorithms.sa import simulated_annealing
from mlrose_hiive.algorithms.ga import  genetic_alg

from mlrose_hiive.neural._nn_base import _NNBase
from mlrose_hiive.neural.activation import (identity, relu, sigmoid, tanh)
from mlrose_hiive.neural.utils.weights import gradient_descent_original


class _NNCore(_NNBase):
    """Core class for neural networks.

    Warning: This class should not be used directly.
    Use derived classes instead.
    """

    @abstractmethod
    def __init__(self, hidden_nodes=None, activation='relu', algorithm='random_hill_climb', max_iters=100, bias=True,
                 is_classifier=True, learning_rate=0.1, early_stopping=False, clip_max=1e+10, restarts=0,
                 schedule=GeomDecay(), pop_size=200, mutation_prob=0.1, max_attempts=10, random_state=None,
                 curve=False):

        super().__init__()
        if hidden_nodes is None:
            self.hidden_nodes = []
        else:
            self.hidden_nodes = hidden_nodes

        self.activation_dict = {'identity': identity,
                                'relu': relu,
                                'sigmoid': sigmoid,
                                'tanh': tanh}
        self.activation = activation
        self.algorithm = algorithm
        self.max_iters = max_iters
        self.bias = bias
        self.is_classifier = is_classifier
        self.learning_rate = learning_rate
        self.early_stopping = early_stopping
        self.clip_max = clip_max
        self.restarts = restarts
        self.schedule = schedule
        self.pop_size = pop_size
        self.mutation_prob = mutation_prob
        self.max_attempts = max_attempts
        self.random_state = random_state
        self.curve = curve

        self.node_list = []
        self.fitted_weights = []
        self.loss = np.inf
        self.output_activation = None
        self.predicted_probs = []
        self.fitness_curve = []

    def _validate(self):
        if (not isinstance(self.max_iters, int) and self.max_iters != np.inf
            and not self.max_iters.is_integer()) or (self.max_iters < 0):
            raise Exception("""max_iters must be a positive integer.""")

        if not isinstance(self.bias, bool):
            raise Exception("""bias must be True or False.""")

        if not isinstance(self.is_classifier, bool):
            raise Exception("""is_classifier must be True or False.""")

        if self.learning_rate <= 0:
            raise Exception("""learning_rate must be greater than 0.""")

        if not isinstance(self.early_stopping, bool):
            raise Exception("""early_stopping must be True or False.""")

        if self.clip_max <= 0:
            raise Exception("""clip_max must be greater than 0.""")

        if (not isinstance(self.max_attempts, int) and not
        self.max_attempts.is_integer()) or (self.max_attempts < 0):
            raise Exception("""max_attempts must be a positive integer.""")

        if self.pop_size < 0:
            raise Exception("""pop_size must be a positive integer.""")
        elif not isinstance(self.pop_size, int):
            if self.pop_size.is_integer():
                self.pop_size = int(self.pop_size)
            else:
                raise Exception("""pop_size must be a positive integer.""")

        if (self.mutation_prob < 0) or (self.mutation_prob > 1):
            raise Exception("""mutation_prob must be between 0 and 1.""")

        if self.activation is None or \
                self.activation not in self.activation_dict.keys():
            raise Exception("""Activation function must be one of: 'identity',
                    'relu', 'sigmoid' or 'tanh'.""")

        if self.algorithm not in ['random_hill_climb', 'simulated_annealing',
                                  'genetic_alg', 'gradient_descent']:
            raise Exception("""Algorithm must be one of: 'random_hill_climb',
                    'simulated_annealing', 'genetic_alg',
                    'gradient_descent'.""")

    def fit(self, X, y=None, init_weights=None):
        """Fit neural network to data.

        Parameters
        ----------
        X: array
            Numpy array containing feature dataset with each row
            representing a single observation.

        y: array
            Numpy array containing data labels. Length must be same as
            length of X.

        init_state: array, default: None
            Numpy array containing starting weights for algorithm.
            If :code:`None`, then a random state is used.
        """
        self._validate()

        X, y = self._format_x_y_data(X, y)

        node_list = self._build_node_list(X, y, self.hidden_nodes, self.bias)

        num_nodes = self._calculate_state_size(node_list)

        if init_weights is not None and len(init_weights) != num_nodes:
            raise Exception("""init_weights must be None or have length %d"""
                            % (num_nodes,))

        # Set random seed
        if isinstance(self.random_state, int) and self.random_state > 0:
            np.random.seed(self.random_state)

        fitness, problem = self._build_problem_and_fitness_function(X, y,
                                                                    node_list,
                                                                    self.activation_dict[self.activation],
                                                                    self.learning_rate,
                                                                    self.bias,
                                                                    self.clip_max,
                                                                    self.is_classifier)

        if self.algorithm == 'random_hill_climb':
            fitness_curve, fitted_weights, loss = self.__run_with_rhc(init_weights, num_nodes, problem)

        elif self.algorithm == 'simulated_annealing':
            fitness_curve, fitted_weights, loss = self._run_with_sa(init_weights, num_nodes, problem)
        elif self.algorithm == 'genetic_alg':
            fitness_curve, fitted_weights, loss = self._run_with_ga(problem)
        else:  # Gradient descent case
            fitness_curve, fitted_weights, loss = self._run_with_gd(init_weights, num_nodes, problem)

        # Save fitted weights and node list
        self.node_list = node_list
        self.fitted_weights = fitted_weights
        self.loss = loss
        self.output_activation = fitness.get_output_activation()

        if self.curve:
            self.fitness_curve = fitness_curve

        return self

    def _run_with_gd(self, init_weights, num_nodes, problem):
        if init_weights is None:
            init_weights = np.random.uniform(-1, 1, num_nodes)

        fitted_weights, loss, fitness_curve = gradient_descent_original(
            problem,
            max_attempts=self.max_attempts if self.early_stopping else self.max_iters,
            max_iters=self.max_iters,
            curve=self.curve,
            init_state=init_weights)

        return ([] if fitness_curve is None else fitness_curve), fitted_weights, loss

    def _run_with_ga(self, problem):
        fitness_curve = []
        if self.curve:
            fitted_weights, loss, fitness_curve = genetic_alg(
                problem,
                pop_size=self.pop_size,
                mutation_prob=self.mutation_prob,
                max_attempts=self.max_attempts if self.early_stopping else
                self.max_iters,
                max_iters=self.max_iters,
                curve=self.curve)
        else:
            fitted_weights, loss, _ = genetic_alg(
                problem,
                pop_size=self.pop_size, mutation_prob=self.mutation_prob,
                max_attempts=self.max_attempts if self.early_stopping else
                self.max_iters,
                max_iters=self.max_iters,
                curve=self.curve)
        return fitness_curve, fitted_weights, loss

    def _run_with_sa(self, init_weights, num_nodes, problem):
        fitness_curve = []
        if init_weights is None:
            init_weights = np.random.uniform(-1, 1, num_nodes)
        if self.curve:
            fitted_weights, loss, fitness_curve = simulated_annealing(
                problem,
                schedule=self.schedule,
                max_attempts=self.max_attempts if self.early_stopping else
                self.max_iters,
                max_iters=self.max_iters,
                init_state=init_weights,
                curve=self.curve)
        else:
            fitted_weights, loss, _ = simulated_annealing(
                problem,
                schedule=self.schedule,
                max_attempts=self.max_attempts if self.early_stopping else
                self.max_iters,
                max_iters=self.max_iters,
                init_state=init_weights,
                curve=self.curve)
        return fitness_curve, fitted_weights, loss

    def __run_with_rhc(self, init_weights, num_nodes, problem):
        fitness_curve = []
        fitted_weights = []
        loss = np.inf
        # Can't use restart feature of random_hill_climb function, since
        # want to keep initial weights in the range -1 to 1.
        for _ in range(self.restarts + 1):
            if init_weights is None:
                init_weights = np.random.uniform(-1, 1, num_nodes)

            if self.curve:
                current_weights, current_loss, fitness_curve = \
                    random_hill_climb(problem,
                                      max_attempts=self.max_attempts if
                                      self.early_stopping else
                                      self.max_iters,
                                      max_iters=self.max_iters,
                                      restarts=0, init_state=init_weights,
                                      curve=self.curve)
            else:
                current_weights, current_loss, _ = random_hill_climb(
                    problem,
                    max_attempts=self.max_attempts if self.early_stopping
                    else self.max_iters,
                    max_iters=self.max_iters,
                    restarts=0, init_state=init_weights, curve=self.curve)

            if current_loss < loss:
                fitted_weights = current_weights
                loss = current_loss
        return fitness_curve, fitted_weights, loss

    def predict(self, X):
        """Use model to predict data labels for given feature array.

        Parameters
        ----------
        X: array
            Numpy array containing feature dataset with each row
            representing a single observation.

        Returns
        -------
        y_pred: array
            Numpy array containing predicted data labels.
        """
        if not np.shape(X)[1] == (self.node_list[0] - self.bias):
            raise Exception("""The number of columns in X must equal %d"""
                            % ((self.node_list[0] - self.bias),))

        y_pred, pp = self._predict(X=X,
                                   fitted_weights=self.fitted_weights,
                                   node_list=self.node_list,
                                   input_activation=self.activation_dict[self.activation],
                                   output_activation=self.output_activation,
                                   bias=self.bias,
                                   is_classifier=self.is_classifier)
        self.predicted_probs = pp
        return y_pred



""" Classes for defining optimization problem objects."""

# Author: Genevieve Hayes
# License: BSD 3 clause

from .arith_decay import ArithDecay
from .geom_decay import GeomDecay
from .exp_decay import ExpDecay
from .custom_schedule import CustomSchedule



""" GA Mutators."""

# Author: Genevieve Hayes
# Modified: Andrew Rollings
# License: BSD 3 clause
from abc import ABC, abstractmethod


class _MutatorBase(ABC):
    def __init__(self, opt_prob):
        self._opt_prob = opt_prob
        self._length = opt_prob.length

    @abstractmethod
    def mutate(self, child, mutation_probability):
        pass



""" Functions to implement the randomized optimization and search algorithms.
"""

# Author: Genevieve Hayes (modified by Andrew Rollings)
# License: BSD 3 clause

from .short_name_decorator import short_name, get_short_name



""" Classes for defining fitness functions."""

# Author: Genevieve Hayes (Modified by Andrew Rollings)
# License: BSD 3 clause

import numpy as np


class Queens:
    """Fitness function for N-Queens optimization problem. Evaluates the
    fitness of an n-dimensional state vector
    :math:`x = [x_{0}, x_{1}, \\ldots, x_{n-1}]`, where :math:`x_{i}`
    represents the row position (between 0 and n-1, inclusive) of the 'queen'
    in column i, as the number of pairs of attacking queens.

    Example
    -------
    .. highlight:: python
    .. code-block:: python

        >>> import mlrose_hiive
        >>> import numpy as np
        >>> fitness = mlrose_hiive.Queens()
        >>> state = np.array([1, 4, 1, 3, 5, 5, 2, 7])
        >>> fitness.evaluate(state)
        6

    References
    ----------
    Russell, S. and P. Norvig (2010). *Artificial Intelligence: A Modern
    Approach*, 3rd edition. Prentice Hall, New Jersey, USA.

    Note
    ----
    The Queens fitness function is suitable for use in discrete-state
    optimization problems *only*.
    """

    def __init__(self):

        self.prob_type = 'discrete'

    @staticmethod
    def shift(a, num, fill_value=np.nan):
        result = np.empty(a.shape)
        if num > 0:
            result[:num] = fill_value
            result[num:] = a[:-num]
        elif num < 0:
            result[num:] = fill_value
            result[:num] = a[-num:]
        else:
            result[:] = a
        return result

    def evaluate(self, state):
        """Evaluate the fitness of a state vector.

        Parameters
        ----------
        state: array
            State array for evaluation.

        Returns
        -------
        fitness: float
            Value of fitness function.
        """

        # check for horizontal matches.
        f_h = (np.unique(state, return_counts=True)[1]-1).sum()

        # check for diagonal matches.
        # look at the state_shifts to figure out how this works. (I'm quite pleased with it)
        ls = state.size
        # rows 0-3:   checking up left.
        # rows 4-7:   checking down right.
        # rows 8-11:  checking up right
        # rows 12-15: checking down left
        state_shifts = np.array([self.shift(state, i)+i for i in np.arange(1-ls, ls) if i != 0] +
                                [self.shift(state, -i)+i for i in np.arange(1-ls, ls) if i != 0])
        # state_shifts[(state_shifts < 0)] = np.NaN
        # state_shifts[(state_shifts >= ls)] = np.NaN

        f_d = np.sum(state_shifts == state) // 2  # each diagonal piece is counted twice
        fitness = f_h + f_d
        return fitness

    def get_prob_type(self):
        """ Return the problem type.

        Returns
        -------
        self.prob_type: string
            Specifies problem type as 'discrete', 'continuous', 'tsp'
            or 'either'.
        """
        return self.prob_type



""" MLROSe setup file."""

# Author: Genevieve Hayes
# Modified: Andrew Rollings
# License: BSD 3 clause

from setuptools import setup

def readme():
    """
    Function to read the long description for the MLROSe package.
    """
    with open('README.md') as _file:
        return _file.read()

VERSION = '2.2.0'

setup(name='mlrose_hiive',
      version=VERSION,
      description="MLROSe: Machine Learning, Randomized Optimization and"
      + " Search (hiive extended remix)",
      long_description=readme(),
      long_description_content_type='text/markdown',
      url='https://github.com/hiive/mlrose',
      author='Genevieve Hayes (modified by Andrew Rollings)',
      license='BSD',
      download_url='https://github.com/hiive/mlrose/archive/'+VERSION+'.tar.gz',
      classifiers=[
          "Intended Audience :: Education",
          "Intended Audience :: Science/Research",
          "Natural Language :: English",
          "License :: OSI Approved :: BSD License",
          "Programming Language :: Python",
          "Programming Language :: Python :: 3",
          "Topic :: Scientific/Engineering",
          "Topic :: Scientific/Engineering :: Artificial Intelligence",
          "Topic :: Scientific/Engineering :: Mathematics",
          "Topic :: Software Development :: Libraries",
          "Topic :: Software Development :: Libraries :: Python Modules"],
      packages=['mlrose_hiive','mlrose_hiive.runners','mlrose_hiive.generators', 'mlrose_hiive.algorithms',
                'mlrose_hiive.algorithms.decay', 'mlrose_hiive.algorithms.crossovers',
                'mlrose_hiive.opt_probs', 'mlrose_hiive.fitness', 'mlrose_hiive.algorithms.mutators',
                'mlrose_hiive.neural', 'mlrose_hiive.neural.activation', 'mlrose_hiive.neural.fitness',
                'mlrose_hiive.neural.utils', 'mlrose_hiive.decorators',
                'mlrose_hiive.gridsearch'],
      install_requires=['numpy', 'scipy', 'scikit-learn', 'pandas', 'networkx', 'joblib'],
      python_requires='>=3',
      zip_safe=False)



import mlrose_hiive
from mlrose_hiive.decorators import short_name
from mlrose_hiive.runners._runner_base import _RunnerBase

"""
Example usage:

    experiment_name = 'example_experiment'
    problem = TSPGenerator.generate(seed=SEED, number_of_cities=22)

    mmc = MIMICRunner(problem=problem,
                      experiment_name=experiment_name,
                      output_directory=OUTPUT_DIRECTORY,
                      seed=SEED,
                      iteration_list=2 ** np.arange(10),
                      max_attempts=500,
                      keep_percent_list=[0.25, 0.5, 0.75])
                      
    # the two data frames will contain the results
    df_run_stats, df_run_curves = mmc.run()
"""


@short_name('mimic')
class MIMICRunner(_RunnerBase):

    def __init__(self, problem, experiment_name, seed, iteration_list, population_sizes,
                 keep_percent_list, max_attempts=500, generate_curves=True, use_fast_mimic=False, **kwargs):
        super().__init__(problem=problem, experiment_name=experiment_name, seed=seed, iteration_list=iteration_list,
                         max_attempts=max_attempts, generate_curves=generate_curves,
                         **kwargs)
        self.keep_percent_list = keep_percent_list
        self.population_sizes = population_sizes
        self._use_fast_mimic = None
        if hasattr(problem, 'set_mimic_fast_mode') and callable(getattr(problem, 'set_mimic_fast_mode')):
            self._use_fast_mimic = use_fast_mimic
            problem.set_mimic_fast_mode(use_fast_mimic)

    def _setup(self):
        super()._setup()
        if self._use_fast_mimic is not None:
            self._log_current_argument('use_fast_mimic', self._use_fast_mimic)

    def run(self):
        return super().run_experiment_(algorithm=mlrose_hiive.mimic,
                                       pop_size=('Population Size', self.population_sizes),
                                       keep_pct=('Keep Percent', self.keep_percent_list))



from mlrose_hiive.opt_probs import MaxKColorOpt
import numpy as np
import networkx as nx


class MaxKColorGenerator:
    @staticmethod
    def generate(seed, number_of_nodes=20, max_connections_per_node=4, max_colors=None):

        """
        >>> edges = [(0, 1), (0, 2), (0, 4), (1, 3), (2, 0), (2, 3), (3, 4)]
        >>> fitness = mlrose_hiive.MaxKColor(edges)
        >>> state = np.array([0, 1, 0, 1, 1])
        >>> fitness.evaluate(state)
        """
        np.random.seed(seed)
        # all nodes have to be connected, somehow.
        node_connection_counts = 1 + np.random.randint(max_connections_per_node, size=number_of_nodes)

        node_connections = {}
        nodes = range(number_of_nodes)
        for n in nodes:
            all_other_valid_nodes = [o for o in nodes if (o != n and (o not in node_connections or
                                                                      n not in node_connections[o]))]
            count = min(node_connection_counts[n], len(all_other_valid_nodes))
            other_nodes = sorted(np.random.choice(all_other_valid_nodes, count, replace=False))
            node_connections[n] = [(n, o) for o in other_nodes]

        # check connectivity
        g = nx.Graph()
        g.add_edges_from([x for y in node_connections.values() for x in y])

        for n in nodes:
            cannot_reach = [(n, o) if n < o else (o, n) for o in nodes if o not in nx.bfs_tree(g, n).nodes()]
            for s, f in cannot_reach:
                g.add_edge(s, f)
                check_reach = len([(n, o) if n < o else (o, n) for o in nodes if o not in nx.bfs_tree(g, n).nodes()])
                if check_reach == 0:
                    break

        edges = [(s, f) for (s, f) in g.edges()]
        problem = MaxKColorOpt(edges=edges, length=number_of_nodes, max_colors=max_colors, source_graph=g)
        return problem



""" Crossover implementations for GA.
"""

# Author: Genevieve Hayes (modified by Andrew Rollings)
# License: BSD 3 clause


from abc import ABC, abstractmethod


class _CrossOverBase(ABC):
    def __init__(self, opt_prob):
        self._opt_prob = opt_prob
        self._length = opt_prob.length

    @abstractmethod
    def mate(self, p1, p2):
        pass



""" Classes for defining neural network weight optimization problems."""

# Author: Genevieve Hayes (modified by Andrew Rollings)
# License: BSD 3 clause

from .identity import (identity)
from .relu import (relu)
from .sigmoid import (sigmoid)
from .softmax import (softmax)
from .tanh import (tanh)



""" MLROSe initialization file."""

# Author: Genevieve Hayes (modified by Andrew Rollings)
# License: BSD 3 clause

from .algorithms.ga import (genetic_alg)
from .algorithms.sa import (simulated_annealing)
from .algorithms.hc import (hill_climb)
from .algorithms.rhc import (random_hill_climb)
from .algorithms.gd import (gradient_descent)
from .algorithms.mimic import (mimic)
from .algorithms.decay import GeomDecay, ArithDecay, ExpDecay, CustomSchedule
from .algorithms.crossovers import OnePointCrossOver, UniformCrossOver, TSPCrossOver
from .algorithms.mutators import ChangeOneMutator, DiscreteMutator, SwapMutator, ShiftOneMutator
from .fitness import (OneMax, FlipFlop, FourPeaks, SixPeaks, ContinuousPeaks,
                      Knapsack, TravellingSales, Queens, MaxKColor,
                      CustomFitness)
from .neural import NeuralNetwork, LinearRegression, LogisticRegression, _nn_core, NNClassifier
from .neural.activation import (identity, relu, sigmoid, softmax, tanh)
from .neural.fitness import NetworkWeights
from .neural.utils.weights import (flatten_weights, unflatten_weights)

from .gridsearch import GridSearchMixin

from .opt_probs import DiscreteOpt, ContinuousOpt, KnapsackOpt, TSPOpt, QueensOpt, FlipFlopOpt, MaxKColorOpt

from .runners import GARunner, MIMICRunner, RHCRunner, SARunner, NNGSRunner
from .runners import (build_data_filename)
from .generators import (MaxKColorGenerator, QueensGenerator, FlipFlopGenerator, TSPGenerator, KnapsackGenerator,
                         ContinuousPeaksGenerator)



""" Classes for defining fitness functions."""

# Author: Genevieve Hayes (Modified by Andrew Rollings)
# License: BSD 3 clause

import numpy as np


class CustomFitness:
    """Class for generating your own fitness function.

    Parameters
    ----------
    fitness_fn: callable
        Function for calculating fitness of a state with the signature
        :code:`fitness_fn(state, **kwargs)`.

    problem_type: string, default: 'either'
        Specifies problem type as 'discrete', 'continuous', 'tsp' or 'either'
        (denoting either discrete or continuous).

    kwargs: additional arguments
        Additional parameters to be passed to the fitness function.

    Example
    -------
    .. highlight:: python
    .. code-block:: python

        >>> import mlrose_hiive
        >>> import numpy as np
        >>> def cust_fn(state, c): return c*np.sum(state)
        >>> kwargs = {'c': 10}
        >>> fitness = mlrose_hiive.CustomFitness(cust_fn, **kwargs)
        >>> state = np.array([1, 2, 3, 4, 5])
        >>> fitness.evaluate(state)
        150
    """

    def __init__(self, fitness_fn, problem_type='either', **kwargs):

        if problem_type not in ['discrete', 'continuous', 'tsp', 'either']:
            raise Exception("""problem_type does not exist.""")
        self.fitness_fn = fitness_fn
        self.problem_type = problem_type
        self.kwargs = kwargs

    def evaluate(self, state):
        """Evaluate the fitness of a state vector.

        Parameters
        ----------
        state: array
            State array for evaluation.

        Returns
        -------
        fitness: float
            Value of fitness function.
        """

        fitness = self.fitness_fn(state, **self.kwargs)
        return fitness

    def get_prob_type(self):
        """ Return the problem type.

        Returns
        -------
        self.prob_type: string
            Specifies problem type as 'discrete', 'continuous', 'tsp'
            or 'either'.
        """
        return self.problem_type



"""
TSP Crossover implementation for GA.
"""

# Author: Genevieve Hayes (modified by Andrew Rollings)
# License: BSD 3 clause

import numpy as np

from mlrose_hiive.algorithms.crossovers._crossover_base import _CrossOverBase


class TSPCrossOver(_CrossOverBase):

    def __init__(self, opt_prob):
        super().__init__(opt_prob)

    def mate(self, p1, p2):
        return self._mate_fill(p1, p2)
        """
        if np.random.randint(2) == 0:
            return self._mate_fill(p1, p2)
        else:
            return self._mate_traverse(p1, p2)
        """

    def _mate_fill(self, p1, p2):
        if self._length > 1:
            n = 1 + np.random.randint(self._length - 1)
            child = np.array([0] * self._length)
            child[:n] = p1[:n]

            unvisited = [node for node in p2 if node not in p1[:n]]
            child[n:] = unvisited
        elif np.random.randint(2) == 0:
            child = np.copy(p1)
        else:
            child = np.copy(p2)
        return child

    def _mate_traverse(self, parent_1, parent_2):
        if self._length > 1:
            next_a = np.append(parent_1[1:], parent_1[-1])
            next_b = np.append(parent_2[1:], parent_2[-1])

            visited = [False] * self._length
            child = np.array([0] * self._length)

            v = np.random.randint(len(parent_1))
            child[0] = v
            visited[v] = True
            for i in range(1, len(child)):
                cur = child[i-1]
                na = next_a[cur]
                nb = next_b[cur]
                va = visited[na]
                vb = visited[nb]
                if va and not vb:
                    nx = nb
                elif not va and vb:
                    nx = na
                elif not va and not vb:
                    fa = self._opt_prob.fitness_fn.calculate_fitness([cur, na])
                    fb = self._opt_prob.fitness_fn.calculate_fitness([cur, nb])
                    nx = nb if fa > fb else na  # opposite because they're distance and smaller is better
                else:
                    while True:
                        nx = np.random.randint(len(parent_1))
                        if not visited[nx]:
                            break
                child[i] = nx
                visited[nx] = True
        elif np.random.randint(2) == 0:
            child = np.copy(parent_1, copy=True)
        else:
            child = np.copy(parent_2, copy=True)
        return child



import sklearn.metrics as skmt
import sklearn.model_selection as skms
import inspect


class GridSearchMixin:
    def __init__(self, scorer_method=None):
        self._scorer_method = skmt.balanced_accuracy_score if scorer_method is None else scorer_method
        self._params = inspect.signature(self._scorer_method)
        self._get_y_argmax = False

    def _perform_grid_search(self, classifier, x_train, y_train, cv, parameters, n_jobs=1, verbose=False):
        scorer = self.make_scorer()
        search_results = skms.GridSearchCV(classifier,
                                           parameters,
                                           cv=cv,
                                           scoring=scorer,
                                           n_jobs=n_jobs,
                                           return_train_score=True,
                                           verbose=verbose)
        search_results.fit(x_train, y_train)
        return search_results

    def make_scorer(self):
        scorer = skmt.make_scorer(self._grid_search_score_intercept)
        return scorer

    def score(self, **kwargs):
        score = self._grid_search_score_intercept(**kwargs)
        return score

    def _grid_search_score_intercept(self, y_pred, y_true, **kwargs):
        cleaned_kwargs = {k: v for k, v in kwargs.items() if k in list(inspect.signature(self._scorer_method).parameters)}

        # workaround for odd usage from sklearn
        if not hasattr(self, '_get_y_argmax'):
            self._get_y_argmax = False

        if not self._get_y_argmax and len(y_pred.shape) > 1 and len(y_true.shape) > 1:
            try:
                return self._scorer_method(y_pred=y_pred, y_true=y_true, **cleaned_kwargs)
            except:
                self._get_y_argmax = True

        if self._get_y_argmax:
            y_pred = y_pred.argmax(axis=1)
            y_true = y_true.argmax(axis=1)

        try:
            return self._scorer_method(y_pred=y_pred, y_true=y_true, **cleaned_kwargs)
        except:
            return self._scorer_method(y_true=y_true, **cleaned_kwargs)




""" Classes for defining fitness functions."""

# Author: Genevieve Hayes (Modified by Andrew Rollings)
# License: BSD 3 clause

import numpy as np


class ContinuousPeaks:
    """Fitness function for Continuous Peaks optimization problem. Evaluates
    the fitness of an n-dimensional state vector :math:`x`, given parameter T,
    as:

    .. math::
        Fitness(x, T) = \\max(max\\_run(0, x), max\\_run(1, x)) + R(x, T)

    where:

    * :math:`max\\_run(b, x)` is the length of the maximum run of b's
      in :math:`x`;
    * :math:`R(x, T) = n`, if (:math:`max\\_run(0, x) > T` and
      :math:`max\\_run(1, x) > T`); and
    * :math:`R(x, T) = 0`, otherwise.

    Parameters
    ----------
    t_pct: float, default: 0.1
        Threshold parameter (T) for Continuous Peaks fitness function,
        expressed as a percentage of the state space dimension, n (i.e.
        :math:`T = t_{pct} \\times n`).

    Example
    -------
    .. highlight:: python
    .. code-block:: python

        >>> import mlrose_hiive
        >>> import numpy as np
        >>> fitness = mlrose_hiive.ContinuousPeaks(t_pct=0.15)
        >>> state = np.array([0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1])
        >>> fitness.evaluate(state)
        17

    Note
    ----
    The Continuous Peaks fitness function is suitable for use in bit-string
    (discrete-state with :code:`max_val = 2`) optimization problems *only*.
    """

    def __init__(self, t_pct=0.1):

        self.t_pct = t_pct
        self.prob_type = 'discrete'

        if (self.t_pct < 0) or (self.t_pct > 1):
            raise Exception("""t_pct must be between 0 and 1.""")

    def evaluate(self, state):
        """Evaluate the fitness of a state vector.

        Parameters
        ----------
        state: array
            State array for evaluation.

        Returns
        -------
        fitness: float
            Value of fitness function.
        """
        _n = len(state)
        _t = np.ceil(self.t_pct*_n)

        # Calculate length of maximum runs of 0's and 1's
        max_0 = self.max_run(0, state)
        max_1 = self.max_run(1, state)

        # Calculate R(X, T)
        if (max_0 > _t and max_1 > _t):
            _r = _n
        else:
            _r = 0

        # Evaluate function
        fitness = max(max_0, max_1) + _r

        return fitness

    def get_prob_type(self):
        """ Return the problem type.

        Returns
        -------
        self.prob_type: string
            Specifies problem type as 'discrete', 'continuous', 'tsp'
            or 'either'.
        """
        return self.prob_type

    @staticmethod
    def max_run(_b, _x):
        """Determine the length of the maximum run of b's in vector x.

        Parameters
        ----------
        b: int
            Integer for counting.

        x: array
            Vector of integers.

        Returns
        -------
        max: int
            Length of maximum run of b's.
        """
        # Initialize counter
        _max = 0
        run = 0

        # Iterate through values in vector
        for i in _x:
            if i == _b:
                run += 1
            else:
                if run > _max:
                    _max = run

                run = 0

        if (_x[-1] == _b) and (run > _max):
            _max = run

        return _max



""" GA Mutators."""

# Author: Genevieve Hayes
# Modified: Andrew Rollings
# License: BSD 3 clause
import numpy as np

from mlrose_hiive.algorithms.mutators._mutator_base import _MutatorBase


class SwapMutator(_MutatorBase):

    def __init__(self, opt_prob):
        super().__init__(opt_prob)

    def mutate(self, child, mutation_probability):
        if np.random.rand() < mutation_probability:
            # do swap mutation
            m1 = np.random.randint(len(child))
            m2 = np.random.randint(len(child))
            tmp = child[m1]
            child[m1] = child[m2]
            child[m2] = tmp
        return child
